DebuffFilter = LibStub("AceAddon-3.0"):NewAddon("DebuffFilter", "AceConsole-3.0", "AceEvent-3.0")
-- ButtonFacade library skins buttons (can change their display in many different ways)
local LBF = LibStub("LibButtonFacade", true)
-- time left in which tenths of a second are displayed
local DEBUFF_FILTER_TIMELEFT_TENTHSSEC = 5

local DEBUFF_FILTER_NO_COOLDOWN = 0
local DEBUFF_FILTER_COOLDOWN_ONLY = 1
local DEBUFF_FILTER_COOLDOWN_AND_TIMER = 2

local DEBUFF_FILTER_NO_SORT = 0
local DEBUFF_FILTER_SORT_ALPHA_ATOZ = 1
local DEBUFF_FILTER_SORT_ALPHA_ZTOA = 2
local DEBUFF_FILTER_SORT_TIMEREM_LONGTOSHORT = 3
local DEBUFF_FILTER_SORT_TIMEREM_SHORTTOLONG = 4

-- default value to update time display when debuff is about to finish
local DEBUFF_FILTER_TIMELEFT_NUMFRACSEC = 0.2

function DebuffFilter:DebugTrace (arg1, enter)
	if self.db and DebuffFilter.PlayerConfig.General.showTrace == true then
		if (enter) then
			DEFAULT_CHAT_FRAME:AddMessage("DebuffFilter:"..arg1.." enter")
		else
			DEFAULT_CHAT_FRAME:AddMessage("DebuffFilter:"..arg1.." exit")
		end
	end
end

function DebuffFilter:DebugPrint (arg1)
	if DebuffFilter.PlayerConfig.General.showDebug == true then
		DEFAULT_CHAT_FRAME:AddMessage("DebuffFilter:"..arg1)
	end
end

-- grabs settings for the options screen, this is done by traversing
-- the configuration table which mirrors the table used for displaying
-- the options screen
function DebuffFilter:getfunc(info)
	local value = DebuffFilter.PlayerConfig
	for i = 1, #info do
		value = value[info[i]]
	end
	return value
end

-- modify configuration table, by traversing it -- it mirrors the table
-- used for displaying the options screen
function DebuffFilter:setfunc(info, value)
	local parent = DebuffFilter.PlayerConfig

	for i = 1,#info-1 do
		parent = parent[info[i]]
	end
	parent[info[#info]] = value
end

local BuffOrDebuff = {
	type = "group",
	name = function(info) return info[#info] end,
    args = {
		-- only (de)buff player applied is shown
		selfapplied={
			name= DFILTER_OPTIONS_SELFAPPLIED,
			desc= DFILTER_OPTIONS_SELFAPPLIED_TOOLTIP,
			type= "toggle",
			set = function(info, value)
					if not value then value = nil end
					DebuffFilter:setfunc(info,value)
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			get = "getfunc",
			order=5,
		},
		-- dont combine (de)buffs
		dontcombine={
			name= DFILTER_OPTIONS_DONTCOMBINE,
			desc= DFILTER_OPTIONS_DONTCOMBINE_TOOLTIP,
			type= "toggle",
			set = function(info, value)
					if not value then value = nil end
					DebuffFilter:setfunc(info,value)
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			get = "getfunc",
			order=6,
		},
		del={
			name= DFILTER_OPTIONS_DEL,
			type= "execute",
			order=3,
			func=function(info)
					info.options.args.Buffs.args[info[2]].args[info[3]].args.List.args[info[5]] = nil 
					DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].List[info[5]] = nil
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
		},
		intervalFractionSecond = {
			order=7,
			name = "Time updates",
			desc = "Period that time is updated below " .. DEBUFF_FILTER_TIMELEFT_TENTHSSEC .. " seconds",
			type = "range",
			min = 0.1,
			max = 1,
			step = 0.1,
			set = function(info, value)
					if value == DEBUFF_FILTER_TIMELEFT_NUMFRACSEC then value = nil end
					DebuffFilter:setfunc(info,value)
				end,
			get = function(info)
					local value = DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].List[info[5]].intervalFractionSecond
					if value == nil then value = DEBUFF_FILTER_TIMELEFT_NUMFRACSEC end
					return value
				end,
		},
		-- display (de)buff if there is a match of value below with name of (de)buff's texture
		texture={
			type="input",
			name=DFILTER_OPTIONS_TEXTURE,
			set = function(info, value)
					value = value:match("^%s*(.-)%s*$")
					if value:len() == 0 then value = nil end
					DebuffFilter:setfunc(info,value)
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			order=8,
			get = "getfunc",
		},
	},
}

local BuffsOrDebuffsList = {
	name = function(info)
			if info[#info] == "Buffs" then return DFILTER_OPTIONS_BUFFS
			else return DFILTER_OPTIONS_DEBUFFS end end,
	type = "group",
	childGroups = "tab",
	args = {
		enterBuff={
			type="input",
			name=DFILTER_OPTIONS_NAME,
			disabled = function(info) return DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].showAllBuffs or not
					DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].enabled end,
			desc= function(info) return "New "..((info[3]):lower()):sub(1,-2).." to add to list" end,
			order=15,
			set = function(info, value)	
					info.options.args.Buffs.args[info[2]].args[info[3]].args.List.args[value] = BuffOrDebuff
					DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].List[value] = {
						intervalFractionSecond = DebuffFilter.PlayerConfig.Frames[info[2]][info[3]].intervalFractionSecond }
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			get = function(info)
					-- remove old debuffs/buffs in list and add correct ones for the target frame
					-- that is selected.  This is done here since the get function does not work
					-- in a group node.  If it does work, it's because a subnode inherited it.
					local buffList = info.options.args.Buffs.args[info[2]].args[info[3]].args.List.args
					for k, v in pairs(buffList) do
						buffList[k] = nil
					end
					for k, v in pairs(DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].List) do
						buffList[k] = BuffOrDebuff
					end
				end,
		},
		showAllBuffs = {
			type = "toggle",
			name = "Show all",
			desc = function(info) return "Show all "..(info[3]):lower() end,
			disabled = function(info) return not DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].enabled end,
			order=6,
			set = function(info, value) 
					DebuffFilter:setfunc(info, value)
					--DebuffFilter_Frame_Update(info[2], info[3])
				end,	
			get = function(info)
					return DebuffFilter:getfunc(info)
				end,
		},
		showAll = {
			type = "multiselect",
			name = "Show all:",
			order = 8,
			width = "half",
			disabled = function(info) return not DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].enabled
						or DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].showAllBuffs end,
			values = { showAllMyBuffs="My", showAllNonRaiderBuffs="Nonraider", showAllStealableBuffs="Stealable",
				showAllMagicBuffs="Magic", showAllPoisonBuffs="Poison", showAllDiseaseBuffs="Disease",
				showAllCurseBuffs="Curse", },
			set = function(info, key, value)
					DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]][key] = value
					local fc = DebuffFilter.Frames[info[2]][info[3]].ComparisonList
					if value then
						if key == "showAllMyBuffs" then fc[key] = {name="ismine",value=true}
						elseif key == "showAllStealableBuffs" then fc[key] = {name="isStealable",value=1} 
						elseif key == "showAllMagicBuffs" then fc[key] = {name="debufftype",value="Magic"} 
						elseif key == "showAllPoisonBuffs" then fc[key] = {name="debufftype",value="Poison"} 
						elseif key == "showAllDiseaseBuffs" then fc[key] = {name="debufftype",value="Disease"} 
						elseif key == "showAllCurseBuffs" then fc[key] = {name="debufftype",value="Curse"} end
					else
						fc[key] = nil
					end
					--DebuffFilter_Frame_Update(info[2], info[3])
				end,
			get = function(info, key)
					return DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]][key]
				end,
		},
		enabled={
			name= "Show Frame",
			desc= "Show or hide current frame",
			type= "toggle",
			order = 5,
			set = function(info, value) 
					DebuffFilter:setfunc(info, value)
					local eFL = DebuffFilter.enabledFramesList
					if value then
						--DebuffFilter_Frame_Update(info[2],info[3])
						DebuffFilter.Frames[info[2]][info[3]].frame:Show()
						if not eFL[info[2]] then
							eFL[info[2]] = {}
						end
						eFL[info[2]][info[3]] = true
					else
						DebuffFilter.Frames[info[2]][info[3]].frame:Hide()
						eFL[info[2]][info[3]] = nil
						if eFL[info[2]].Buffs == nil and eFL[info[2]].Debuffs == nil then
							eFL[info[2]] = nil
						end
					end
				end,
			get = "getfunc",
		},
		copyFromFrame={
			name= function(info) return "Copy "..(info[3]):lower().." from frame:" end,
			type= "select",
			order = 16,
			values= function(info)
					local table = {}
					for k, v in pairs(DebuffFilter.Frames) do
						if k ~= info[2] then table[k] = k end
					end
					return table
				end,
			set= function(info, value)
					local buffList = info.options.args.Buffs.args[info[2]].args[info[3]].args.List.args
					for k, v in pairs(DebuffFilter.PlayerConfig.Buffs[value][info[3]].List) do
						DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].List[k] = v
						buffList[k] = BuffOrDebuff
					end
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
		},
		-- combine buffs and debuffs into selected frame
		combineBuffsDebuffsFrames={
			name= function(info) 
					local ret_str = "Add debuffs"
					if info[3] == "Debuffs" then ret_str = "Add buffs" end
					return ret_str
				end,
			desc= function(info) 
					local ret_str
					if info[3] == "Debuffs" then 
						ret_str = "Add '" .. info[2] .. "' buffs to this frame"
					else
						ret_str = "Add '" .. info[2] .. "' debuffs to this frame"
					end
					return ret_str
				end,
			type= "toggle",
			order = 7,
			set = "setfunc",
			get = "getfunc",
		},
		buffDurationToggle={
			name= "Filter using duration",
			desc = function(info) return "Show only selected "..(info[3]):lower().." that last at most "..
				"this number of seconds -- does not apply to "..(info[3]):lower().." listed below" end,
			type= "toggle",
			order = 11,
			set = "setfunc",
			get = "getfunc",
		},	
		buffExpiretimeToggle={
			name= "Filter using expire time",
			desc = function(info) return "Show only selected "..(info[3]):lower().." when they expire "..
				"in less than (or equal to) this number of seconds" end,
			type= "toggle",
			order = 12,
			set = "setfunc",
			get = "getfunc",
		},
		buffDuration = {
			order=13,
			disabled = function(info) return not DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].buffDurationToggle end,
			name = "Duration",
			desc = function(info) return "Show only selected "..(info[3]):lower().." that last at most "..
				"this number of seconds -- does not apply to "..(info[3]):lower().." listed below" end,
			type = "range",
			min = 10,
			max = 130,
			step = 10,
			set = "setfunc",
			get = "getfunc",
		},
		buffExpiretime = {
			order=14,
			disabled = function(info) return not DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].buffExpiretimeToggle end,
			name = "Expire time",
			desc = function(info) return "Show only selected "..(info[3]):lower().." when they expire "..
				"in less than (or equal to) this number of seconds" end,
			type = "range",
			min = 10,
			max = 130,
			step = 10,
			set = "setfunc",
			get = "getfunc",
		},
		List={
			name= function(info) return "List of "..(info[3]):lower()..":  ENLARGE window to see list" end,
			--name = DFILTER_OPTIONS_NAME,
			type = "group",
			disabled = function(info) return DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].showAllBuffs or not
					DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].enabled end,
			args = {
			},
		},	
	},
}

local frameOptions = {
	type = "group",
	name = function(info)
			if info[#info] == "Buffs" then return DFILTER_OPTIONS_BUFFS
			else return DFILTER_OPTIONS_DEBUFFS end end,
	args = {
        grow = {
        	order=1,
        	name = DFILTER_OPTIONS_GROW,
        	desc = DFILTER_OPTIONS_GROW_TOOLTIP,
        	type = "select",
        	values = {
				rightdown = "Right-Down",
				rightup = "Right-Up",
				leftdown = "Left-Down",
				leftup = "Left-Up",
			},
			set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						DebuffFilter_UpdateLayout(info[2],info[3])
				end,	
            get = "getfunc",
        },
        per_row = {
        	order=2,
        	name = DFILTER_OPTIONS_ROW,
        	desc = DFILTER_OPTIONS_ROW_TOOLTIP,
        	type = "range",
        	min = 1,
        	max = 8,
        	step = 1,
			set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						DebuffFilter_UpdateLayout(info[2],info[3])
				end,	
            get = "getfunc",
        },
		scale = {
        	order=3,
			name = "frame scale",
			desc = "Scales the frame only up or down in size",
			type = "range",
			min = 0.5,
			max = 3,
			step = 0.05,
			set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						DebuffFilter:FrameSetScale(DebuffFilter.Frames[info[2]][info[3]].frame,value)
				end,	
			get = "getfunc",
		},
		cooldown={
			name= DFILTER_OPTIONS_COOLDOWNCOUNT,
			desc="If \"No cooldown\" is selected, value from General Settings is used",
			type= "select",
			order = 4,
			values= {[DEBUFF_FILTER_NO_COOLDOWN]="No cooldown",[DEBUFF_FILTER_COOLDOWN_ONLY]=DFILTER_OPTIONS_COOLDOWNCOUNT,
				[DEBUFF_FILTER_COOLDOWN_AND_TIMER]="Cooldown and time"},
			set = function(info, value) 
					DebuffFilter:setfunc(info, value)
					DebuffFilter_UpdateLayout(info[2],info[3])
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			get = "getfunc",
		},
		reverseCooldown={
        	order=5,
			name= "Reverse cooldown",
			desc= "Reverse cooldown display",
			type= "toggle",
			set = function(info, value) 
					DebuffFilter:setfunc(info, value)
					DebuffFilter_UpdateLayout(info[2],info[3])
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			get = "getfunc",
		},
		sort={
			name=function(info) return "Sort "..(info[3]):lower()..":" end,
			desc="If \"No sort\" is selected, value from General Settings is used",
			type= "select",
			order = 6,
			values= {[DEBUFF_FILTER_NO_SORT]="No sort",
				[DEBUFF_FILTER_SORT_ALPHA_ATOZ]="From A to Z",
				[DEBUFF_FILTER_SORT_ALPHA_ZTOA]="From Z to A",
				[DEBUFF_FILTER_SORT_TIMEREM_LONGTOSHORT]="Longest time to shortest",
				[DEBUFF_FILTER_SORT_TIMEREM_SHORTTOLONG]="Shortest time to longest",
			},
			set = function(info, value) 
					DebuffFilter:setfunc(info, value)
					DebuffFilter_UpdateLayout(info[2],info[3])
					--DebuffFilter_Frame_Update(info[2],info[3])
				end,
			get = "getfunc",
		},
		-- period to update time display when debuff is about to finish
		intervalFractionSecond = {
			order=7,
			name = "Time updates",
			desc = "Period that time is updated below " .. DEBUFF_FILTER_TIMELEFT_TENTHSSEC .. " seconds",
			type = "range",
			min = 0.1,
			max = 1,
			step = 0.1,
			set = function(info, value) 
					DebuffFilter:setfunc(info, value)
					if value == DEBUFF_FILTER_TIMELEFT_NUMFRACSEC then value = nil end
					for k, v in pairs(DebuffFilter.PlayerConfig.Buffs[info[2]][info[3]].List) do
						v.intervalFractionSecond = value
					end
				end,	
			get = "getfunc",
		},
	},
}

local frameLayoutTabs = {
	type = "group",
	name = function(info) return info[#info] end,
	childGroups = "tab",
	args = {
		Debuffs = frameOptions,
		Buffs = frameOptions,
	}	
}

local frameBuffsTabs = {
	type = "group",
	name = function(info) return info[#info] end,
	childGroups = "tab",
	args = {
		Debuffs = BuffsOrDebuffsList,
		Buffs = BuffsOrDebuffsList,
	},
}

local newFrameName = {
	type = "group",
	name = function(info) return info[#info] end,
	args={
		del={
			name= DFILTER_OPTIONS_DEL,
			type= "execute",
			func=function(info)
				info.options.args.Frames.args[info[3]] = nil
				info.options.args.Buffs.args[info[3]] = nil
				info.options.args.General.args.listNewFrames.args[info[3]] = nil
				
				if LBF then
					local groupID
					for a, b in ipairs({"Buffs","Debuffs"}) do
						groupID = info[3].." "..b.." Frame"
						DebuffFilter.Frames[info[3]][b].buttonfacadeGroup:Delete()
						DebuffFilter.PlayerConfig.ButtonFacade[groupID] = nil
					end
				end	

				DebuffFilter.Frames[info[3]]["Buffs"].frame:Hide()
				DebuffFilter.Frames[info[3]]["Debuffs"].frame:Hide()
				DebuffFilter.Frames[info[3]] = nil
				
				DebuffFilter.PlayerConfig.Buffs[info[3]] = nil
				DebuffFilter.PlayerConfig.Frames[info[3]] = nil
				
				DebuffFilter.enabledFramesList[info[3]] = nil
			end,
		},
		newFrameTarget={
			type="input",
			name="Target for new frame:",
			desc="target, player, focus, pet, raid1, focustarget...",
			set = function(info, value)
					DebuffFilter.PlayerConfig.Frames[info[3]].frametarget = value
					--DebuffFilter_Frame_Update(info[3],"Buffs")
					--DebuffFilter_Frame_Update(info[3],"Debuffs")
				end,
			get = function(info)
					return DebuffFilter.PlayerConfig.Frames[info[3]].frametarget
				end,
		},
	},	
}

local options = {
	handler = DebuffFilter,
	name="DebuffFilter",
	type = "group",
	args = {

	BlizOptions ={
		type = "group",
		name = "Open Standalone Dialog",
		args={
			version={
				order=1,
				type="description",
				name=function(info) return "version: "..GetAddOnMetadata("DebuffFilter", "Version") end,
			},
			config = {
				order=11,
				type = "execute",
				name = "Standalone Config",
				desc = "Setup addon in a movable standalone dialog",
				func = "OpenConfigDialog"
			},
		},
	},
	General = {
		order = 1,
		type = "group",
		name = "General Settings",
		childGroups = "tab",
		args = {
			version={
				order=1,
				type="description",
				name=function(info) return "version: "..GetAddOnMetadata("DebuffFilter", "Version") end,
			},
			count={
				order=2,
				name= DFILTER_OPTIONS_COUNT,
				desc= DFILTER_OPTIONS_COUNT_TOOLTIP,
				type= "toggle",
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						for k, v in pairs(DebuffFilter.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								if value then 
									DebuffFilter.Frames[k][b].frameCount:Show()
								else
									DebuffFilter.Frames[k][b].frameCount:Hide()
								end
							end
						end	
					end,
				get = "getfunc",
			},
			cooldown={
				name= DFILTER_OPTIONS_COOLDOWNCOUNT,
				desc=DFILTER_OPTIONS_COOLDOWNCOUNT_TOOLTIP,
				type= "select",
				order = 5,
				values= {[DEBUFF_FILTER_NO_COOLDOWN]="No cooldown",[DEBUFF_FILTER_COOLDOWN_ONLY]=DFILTER_OPTIONS_COOLDOWNCOUNT,
					[DEBUFF_FILTER_COOLDOWN_AND_TIMER]="Cooldown and time"},
				set= function(info, value)
						DebuffFilter:setfunc(info, value)
						for k, v in pairs(DebuffFilter.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								DebuffFilter_UpdateLayout(k,b)
								--DebuffFilter_Frame_Update(k,b)
							end
						end
					end,
				get = "getfunc",
			},
			reverseCooldown={
				order=3,
				name= "Reverse cooldown",
				desc= "Reverse cooldown display",
				type= "toggle",
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						for k, v in pairs(DebuffFilter.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								DebuffFilter_UpdateLayout(k,b)
								--DebuffFilter_Frame_Update(k,b)
							end
						end
						frameOptions.args.reverseCooldown.disabled = value
					end,
				get = "getfunc",
			},	
			sort={
				name= "Sort buffs/debuffs:",
				--desc="",
				type= "select",
				order = 6,
				values= {[DEBUFF_FILTER_NO_SORT]="No sort",
					[DEBUFF_FILTER_SORT_ALPHA_ATOZ]="From A to Z",
					[DEBUFF_FILTER_SORT_ALPHA_ZTOA]="From Z to A",
					[DEBUFF_FILTER_SORT_TIMEREM_LONGTOSHORT]="Longest time to shortest",
					[DEBUFF_FILTER_SORT_TIMEREM_SHORTTOLONG]="Shortest time to longest",
				},	
				set= function(info, value)
						DebuffFilter:setfunc(info, value)
						for k, v in pairs(DebuffFilter.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								DebuffFilter_UpdateLayout(k,b)
								--DebuffFilter_Frame_Update(k,b)
							end
						end
					end,
				get = "getfunc",
			},
			combat={
				order=7,
				name= DFILTER_OPTIONS_COMBAT,
				desc= DFILTER_OPTIONS_COMBAT_TOOLTIP,
				type= "toggle",
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						if value then
							DebuffFilter:RegisterEvent("PLAYER_REGEN_DISABLED");
							DebuffFilter:RegisterEvent("PLAYER_REGEN_ENABLED");
							if (not UnitAffectingCombat("player")) then
								DebuffFilterFrame:Hide();
							end
						else
							DebuffFilter:UnregisterEvent("PLAYER_REGEN_DISABLED");
							DebuffFilter:UnregisterEvent("PLAYER_REGEN_ENABLED");
							DebuffFilterFrame:Show();
						end
					end,	
				get = "getfunc",
			},
			tooltips={
				order=8,
				name= DFILTER_OPTIONS_TOOLTIPS,
				desc= DFILTER_OPTIONS_TOOLTIPS_TOOLTIP,
				type= "toggle",
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						-- see lock option below for explanation
						if not value and DebuffFilter.PlayerConfig.General.lock then 
							DebuffFilter_LockFrames(true) 
						else
							DebuffFilter_LockFrames(false)
						end	
					end,	
				get = "getfunc",
			},
			backdrop={
				order=9,
				name= DFILTER_OPTIONS_BACKDROP,
				desc= DFILTER_OPTIONS_BACKDROP_TOOLTIP,
				type= "toggle",
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						for k, v in pairs(DebuffFilter.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								if value then
									DebuffFilter.Frames[k][b].frameBackdrop:Show()
								else
									DebuffFilter.Frames[k][b].frameBackdrop:Hide()
								end
							end
						end	
					end,	
				get = "getfunc",
			},
			lock={
				order=10,
				name= DFILTER_OPTIONS_LOCK,
				desc= "If tooltips are not shown, mouse click-thru is allowed",
				type= "toggle",
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						-- to allow tooltips to be shown, the frames won't be allowed to be moved, but
						-- the buttons will still accept mouse input
						if value and not DebuffFilter.PlayerConfig.General.tooltips then 
							DebuffFilter_LockFrames(true) 
						else
							DebuffFilter_LockFrames(false)
						end	
					end,	
				get = "getfunc",
			},
			resetposition={
				order=13,
				name= "Reset Frame Positions",
				type= "execute",
				func=function()
					DebuffFilter.xvalues.xvalue = 478
					DebuffFilter.xvalues.nextxvalue = 578
					DebuffFilter.yvalue = 335
					local anchor;
					-- a frame's position is affected by its scale, so I need to reset scales too
					DebuffFilter.PlayerConfig.General.scale = 1.2
					DebuffFilterFrame:SetScale(1.2)
					for k, v in pairs(DebuffFilter.Frames) do
						for a, b in ipairs({"Buffs","Debuffs"}) do
							DebuffFilter.PlayerConfig.Frames[k][b].scale = 1
							DebuffFilter:FrameSetScale(DebuffFilter.Frames[k][b].frame, 1);
							
							anchor = {"TOPLEFT","UIParent","BOTTOMLEFT",DebuffFilter.xvalues.xvalue,DebuffFilter.yvalue}
							DebuffFilter.PlayerConfig.Frames[k][b].anchor = anchor			
							DebuffFilter.Frames[k][b].frame:ClearAllPoints()
							DebuffFilter.Frames[k][b].frame:SetPoint(unpack(anchor))

							DebuffFilter:swapXvalues()

						end
						DebuffFilter:adjustAnchorPoints()
					end
				end,		
			},
			intervalFractionSecond = {
				order=11,
				name = "Time updates",
				desc = "Period that time is updated below " .. DEBUFF_FILTER_TIMELEFT_TENTHSSEC .. " seconds",
				type = "range",
				min = 0.1,
				max = 1,
				step = 0.1,
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						for k, v in pairs(DebuffFilter.PlayerConfig.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								DebuffFilter.PlayerConfig.Frames[k][b].intervalFractionSecond = value
							end
						end
						-- do not set nil value above since frame layout's get function does not check for nil value
						if value == DEBUFF_FILTER_TIMELEFT_NUMFRACSEC then value = nil end
						for k, v in pairs(DebuffFilter.PlayerConfig.Buffs) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								for c, d in pairs(DebuffFilter.PlayerConfig.Buffs[k][b].List) do
									d.intervalFractionSecond = value
								end
							end
						end
					end,	
				get = "getfunc",
			},
			scale = {
				order=12,
				name = DFILTER_OPTIONS_SCALE,
				desc = DFILTER_OPTIONS_SCALE_TOOLTIP,
				type = "range",
				min = 0.5,
				max = 3,
				step = 0.05,
				set = function(info, value) 
						DebuffFilter:setfunc(info, value)
						DebuffFilter:SetScale(value)
					end,	
				get = "getfunc",
			},
			--[[header={
				order=11,
				name="",
				type= "header",
			},]]
			noBoxAroundTime={
				order=14,
				name= "No box around time",
				desc= "No black box around the time",
				type= "toggle",
				set = function(info, value)
						DebuffFilter:setfunc(info, value)
						local alpha = 1
						if value then
							alpha = 0							
						end
						for k, v in pairs(DebuffFilter.Frames) do
							for a, b in ipairs({"Buffs","Debuffs"}) do
								for c, button in pairs(DebuffFilter.Frames[k][b].buttons) do
									button.time_frame.DBFframetexture:SetTexture(0,0,0,alpha)
								end
							end
						end	
					end,
				get = "getfunc",
			},
			newFrame={
				order=15,
				type="input",
				width="double",
				name="Name for new frame:",
				desc="Can duplicate existing frames or show buffs for new targets, like pet, raid1...",
				set = function(info, value)
						info.options.args.Frames.args[value] = frameLayoutTabs
						info.options.args.Buffs.args[value] = frameBuffsTabs
						-- frame's target can be changed by user
						DebuffFilter.PlayerConfig.Frames[value].frametarget = value
						info.options.args.General.args.listNewFrames.args[value] = newFrameName

						local frameitem, groupID
						DebuffFilter.Frames[value] = {}
						for a, b in ipairs({"Buffs","Debuffs"}) do
							DebuffFilter:createFrame(value, b)
							DebuffFilter.PlayerConfig.Frames[value][b].intervalFractionSecond = DebuffFilter.PlayerConfig.General.intervalFractionSecond
							-- if there's ButtonFacade, create a group for frame so it can be skinned
							if LBF then
								groupID = value.." "..b.." Frame"
								frameitem = DebuffFilter.Frames[value][b]
								frameitem.buttonfacadeGroup = LBF:Group("DebuffFilter", groupID)
							end	
						end	
						DebuffFilter:adjustAnchorPoints()
					end,
			},
			listNewFrames={
				order=16,
				type="group",
				name="List of new frames:",
				args={
				},
			},
			showDebug = {
				hidden = true,
				name = "ShowDebug",
				desc = "Show debug messages",
				type = "toggle",
				set = "setfunc",
				get = "getfunc",
			},
			showTrace = {
				hidden = true,
				name = "ShowTrace",
				desc = "Show trace messages",
				type = "toggle",
				set = "setfunc",
				get = "getfunc",
			},
		},	
	},
	Frames = {
		order = 2,
		type = "group",
		name = "Frame Layout",
		childGroups = "select",
		-- tooltip won't appear
		--desc = DFILTER_OPTIONS_TARGET_TOOLTIP,
		args = {
		},
	},
	Buffs = {
		order = 3,
		type = "group",
		name = "Frame Buffs",
		childGroups = "select",
		--desc = DFILTER_OPTIONS_TARGET_TOOLTIP,
		args = {
		},

	},

	},
}

-- taken from bongos, like below
function DebuffFilter:FrameSetScale(frame, scale)
	local ratio, x, y, layout;
	
	ratio = frame:GetScale() / scale;

	x, y = frame:GetLeft() * ratio, frame:GetTop() * ratio;
	-- store new position of frame in config file
	layout = DebuffFilter.PlayerConfig.Frames[frame.DBFframename][frame.DBFbuffOrDebuff]
	layout.anchor = {"TOPLEFT", "UIParent", "BOTTOMLEFT", x, y};			

	frame:ClearAllPoints();
	frame:SetPoint(unpack(layout.anchor));
	frame:SetScale(scale);
end

-- taken from bongos
function DebuffFilter:SetScale(scale)
	local ratio, x, y, frame, layout;

	ratio = DebuffFilterFrame:GetScale() / scale;

	for k, v in pairs(DebuffFilter.Frames) do
		for a, b in ipairs({"Buffs","Debuffs"}) do
			frame = DebuffFilter.Frames[k][b].frame
			x, y = frame:GetLeft() * ratio, frame:GetTop() * ratio;
			-- store new position of frame in config file
			layout = DebuffFilter.PlayerConfig.Frames[k][b]
			layout.anchor = {"TOPLEFT", "UIParent", "BOTTOMLEFT", x, y};			

			frame:ClearAllPoints();
			frame:SetPoint(unpack(layout.anchor));
		end	
	end

	DebuffFilterFrame:SetScale(scale);
end

-- I use two x positions, one for the buff, and the other for the debuff
function DebuffFilter:swapXvalues()
	local tempxvalue
	tempxvalue = DebuffFilter.xvalues.xvalue;
	DebuffFilter.xvalues.xvalue = DebuffFilter.xvalues.nextxvalue;
	DebuffFilter.xvalues.nextxvalue = tempxvalue;
end

-- for new frames, or resetting frames, this function calculates new positions
-- the positions are arranged like a grid
function DebuffFilter:adjustAnchorPoints()
	local yvalue, xvalue, nextxvalue = DebuffFilter.yvalue, DebuffFilter.xvalues.xvalue, DebuffFilter.xvalues.nextxvalue
	yvalue = yvalue - 60;
	-- if y has fallen too much, set its position back to top of screen
	if yvalue < 150 then
		xvalue = nextxvalue + 100
		nextxvalue = xvalue + 100
		yvalue = 515
		-- if x has gone too far right, set its position back to left of screen
		if xvalue > 885 then
			xvalue = xvalue - 1020
			nextxvalue = xvalue + 100
		end
	end
	DebuffFilter.yvalue, DebuffFilter.xvalues.xvalue, DebuffFilter.xvalues.nextxvalue = yvalue, xvalue, nextxvalue
end

-- settings that are stored on file and are referenced through DebuffFilter.PlayerConfig.  
-- ['**'] is a wildcard, and it can be replaced by any string -- it is used by Ace-DB
function DebuffFilter:InitDefaults()
	local defaults = {
		char = {
			General = {
				count = false,
				cooldown = DEBUFF_FILTER_NO_COOLDOWN,
				cooldowncount = false,
				reverseCooldown = false,
				sort = DEBUFF_FILTER_NO_SORT,
				combat = false,
				tooltips = false,
				backdrop = false,
				lock = true,
				scale = 1,
				showDebug = false,
				showTrace = false,
				intervalFractionSecond = DEBUFF_FILTER_TIMELEFT_NUMFRACSEC,
				noBoxAroundTime = false,
			},
			Frames = {
				['**'] = {
					Buffs = {
						per_row = 8,
						scale = 1,
						grow = "rightdown",
						time_tb = "bottom",
						time_lr = "right",
						cooldown = DEBUFF_FILTER_NO_COOLDOWN,
						reverseCooldown = false,
						sort = DEBUFF_FILTER_NO_SORT,
						intervalFractionSecond = DEBUFF_FILTER_TIMELEFT_NUMFRACSEC,
					},
					Debuffs = {
						per_row = 8,
						scale = 1,
						grow = "rightdown",
						time_tb = "bottom",
						time_lr = "right",
						cooldown = DEBUFF_FILTER_NO_COOLDOWN,
						reverseCooldown = false,
						sort = DEBUFF_FILTER_NO_SORT,
						intervalFractionSecond = DEBUFF_FILTER_TIMELEFT_NUMFRACSEC,
					},
				},
			},
			Buffs = {
				target = {
					Debuffs = {
						enabled = true,
						showAllMyBuffs = true,
						showAllNonRaiderBuffs = true,
					},
					Buffs = {
						enabled = true,
						showAllMyBuffs = true,
						showAllNonRaiderBuffs = true,
					},
				},
				player = {
					Debuffs = {
						enabled = true,
						showAllNonRaiderBuffs = true,
					},
					Buffs = {
						enabled = true,
						showAllMyBuffs = true,
					},
				},
				['**'] = {
					Buffs = {
						showAllBuffs = false,
						showAllStealableBuffs = false,
						showAllMagicBuffs = false,
						showAllPoisonBuffs = false,
						showAllDiseaseBuffs = false,
						showAllCurseBuffs = false,
						combineBuffsDebuffsFrames = false,
						buffDurationToggle = false,
						buffExpiretimeToggle = false,
						buffDuration = 30,
						buffExpiretime = 30,
						-- cannot use ['**'] for list of buffs, otherwise buffs with default values won't get stored
						["List"] = {
						},
					},
					Debuffs = {
						showAllBuffs = false,
						showAllStealableBuffs = false,
						showAllMagicBuffs = false,
						showAllPoisonBuffs = false,
						showAllDiseaseBuffs = false,
						showAllCurseBuffs = false,
						combineBuffsDebuffsFrames = false,
						buffDurationToggle = false,
						buffExpiretimeToggle = false,
						buffDuration = 30,
						buffExpiretime = 30,
						["List"] = {
						},
					},
				},
			},
			ButtonFacade = {
			},
		},
	}
	return defaults
end

-- the direction that debuffs/buffs are organized, what side their time is on,
-- and what side the number of debuffs/buffs is placed
DebuffFilter.Orientation = {
	rightdown = { point="LEFT", relpoint="RIGHT", x=4, y=0 },
	rightup = { point="LEFT", relpoint="RIGHT", x=4, y=0 },
	leftdown = { point="RIGHT", relpoint="LEFT", x=-4, y=0 },
	leftup = { point="RIGHT", relpoint="LEFT", x=-4, y=0 },
	bottom = { point="TOP", relpoint="BOTTOM", x=0, y=-2, next_time="top" },
	top = { point="BOTTOM", relpoint="TOP", x=0, y=2, next_time="bottom" },
	left = { point="RIGHT", relpoint="LEFT", x=-4, y=0, next_time="right" },
	right = { point="LEFT", relpoint="RIGHT", x=4, y=0, next_time="left" },
}

function DebuffFilter_Button_OnLoad(self)
	local name = self:GetName();

	self.icon = _G[name .. "Icon"];
	-- frame which holds button's time, and surrounds it with a black background
	self.time_frame = _G[name .. "Duration"]
	self.time_frame.time = _G[name .. "DurationString"];
	self.time_frame.DBFframetexture = _G[name .. "DurationBckgrnd"]
	self.cooldown = _G[name .. "Cooldown"];
	-- current number of debuff/buff's stack
	self.count = _G[name .. "Count"];
	-- number of same debuffs/buffs that are combined
	self.count2 = _G[name .. "Count2"];
	self.border = _G[name .. "Border"];
end

function DebuffFilter_OnMouseDown(self, button)
	if (button == "LeftButton" and IsShiftKeyDown()) then
		-- backdrop has no name, so allow frame to be moved if backdrop is selected
		if self:GetName() == nil or not DebuffFilter.PlayerConfig.General.lock then
			self:GetParent():StartMoving();
		end	
	elseif (button == "RightButton" and IsControlKeyDown()) then
		local next_time;
		local frame = self:GetParent();
		local layout = DebuffFilter.PlayerConfig.Frames[frame.DBFframename][frame.DBFbuffOrDebuff];

		-- switch position of the debuffs/buffs time, if it's bottom make it top
		-- if there's only 1 per row, switch it to left or right side
		if (layout.per_row == 1) then
			next_time = DebuffFilter.Orientation[layout.time_lr].next_time;
			layout.time_lr = next_time;
		else
			next_time = DebuffFilter.Orientation[layout.time_tb].next_time;
			layout.time_tb = next_time;
		end

		-- reposition the times for the debuffs/buffs of a certain frame
		DebuffFilter_SetTimeOrientation(next_time, DebuffFilter.Frames[frame.DBFframename][frame.DBFbuffOrDebuff].buttons);
		DebuffFilter_Print(frame.DBFframename.." "..frame.DBFbuffOrDebuff.. " time orientation: " .. next_time);
	end
end

function DebuffFilter_OnMouseUp(self, button)
	if (button == "LeftButton") then
		local frame = self:GetParent();
		frame:StopMovingOrSizing();
		
		-- store the new frame position in config file
		local anchor = {"TOPLEFT","UIParent","BOTTOMLEFT",frame:GetLeft(),frame:GetTop()};
		DebuffFilter.PlayerConfig.Frames[frame.DBFframename][frame.DBFbuffOrDebuff].anchor = anchor;
	end
end

function DebuffFilter_Print(msg)
	DEFAULT_CHAT_FRAME:AddMessage("|cff00ff00Debuff Filter|r: " .. msg);
end

local hideTime = function(time_frame)
	if time_frame.DBFisTimeVisible then
		--button:SetScript("OnUpdate",nil);
		time_frame:Hide()
		time_frame.DBFisTimeVisible = false
	end	
end

function DebuffFilter_ShowButton(button, bt, framename, buffOrDebuff)

	local time_frame = button.time_frame

	-- sets the color of the border of the debuff, but I can't find
	-- the DebuffTypeColor list anywhere
	if bt.isdebuff then
		local color
		if (bt.debufftype) then
			color = DebuffTypeColor[bt.debufftype];
		else
			color = DebuffTypeColor["none"];
			bt.debufftype = "none"
		end
		-- save time maybe by setting the border color only if it's changed
		if button.DBFdebufftype ~= bt.debufftype then
			button.border:SetVertexColor(color.r, color.g, color.b) 
			button.DBFdebufftype = bt.debufftype
		end	
		button.border:Show()
	else
		button.border:Hide()
	end

	-- needed for gametooltip -- so i know to use SetUnitDebuff or SetUnitBuff
	button.isdebuff = bt.isdebuff
	-- needed for gametooltip, so it knows what (de)buff description to use
	button.index = bt.index;

	if button.DBFnum_combines ~= bt.num_combines then
		button.DBFnum_combines = bt.num_combines
		if bt.num_combines == 1 then
			button.count2:Hide();
		else
			button.count2:SetText(bt.num_combines)
			button.count2:Show();
		end	
	end

	--time_frame.expiretime = bt.expiretime;
	-- save time maybe by setting the texture if it's changed
	if button.DBFtexture ~= bt.texture then
		button.icon:SetTexture(bt.texture);
		button.DBFtexture = bt.texture
	end
	time_frame.intervalFractionSecond = bt.intervalFractionSecond
	-- dont show fractions of a second if 1 is chosen
	time_frame.showFractionSecond = not(time_frame.intervalFractionSecond == 1)

	-- show stack number of a debuff/buff
	if button.DBFapplications ~= bt.applications then
		button.DBFapplications = bt.applications
		if (bt.applications > 1) then
			button.count:SetText(bt.applications);
			button.count:Show();
		else
			button.count:Hide();
		end
	end

	local cooldown = button.cooldown
	local frameitem = DebuffFilter.PlayerConfig.Frames[framename][buffOrDebuff]
	local generaloptions = DebuffFilter.PlayerConfig.General
	-- show time remaining for a debuff/buff, if it exists, as a time
	-- and/or display it as a cooldown on the debuff/buff itself in the form of a rotating shadow
	if (bt.duration and bt.duration > 0) then
		if (frameitem.cooldown > DEBUFF_FILTER_NO_COOLDOWN or generaloptions.cooldown > DEBUFF_FILTER_NO_COOLDOWN) then
			local reverseCooldownState = not(frameitem.reverseCooldown or generaloptions.reverseCooldown)

			-- cannot call setcooldown every 0.025 seconds, otherwise it looks awful, and not like a cooldown
			if not (cooldown.DBFexpiretime == bt.expiretime and cooldown.DBFduration == bt.duration) then
				cooldown:SetCooldown(bt.expiretime-bt.duration, bt.duration)
				cooldown.DBFexpiretime = bt.expiretime
				cooldown.DBFduration = bt.duration
				cooldown:SetReverse(reverseCooldownState)
				cooldown.DBFreverseCooldownState = reverseCooldownState
			else
				cooldown:Show()
				-- setreverse might be an expensive call, so I check a variable to see whether to change it
				if cooldown.DBFreverseCooldownState ~= reverseCooldownState then
					cooldown:SetReverse(reverseCooldownState)
					cooldown.DBFreverseCooldownState = reverseCooldownState
				end	
			end
			
		else
			cooldown:Hide()
		end
		-- make time visible and have it updated every second, if debuff/buff has a time.  Also, the user
		-- must not choose the "cooldown only" option for the frame, or if the "no cooldown" option is chosen
		-- for the frame, the user must not choose the "cooldown only" option in general settings
		if (frameitem.cooldown == DEBUFF_FILTER_COOLDOWN_AND_TIMER or (frameitem.cooldown == DEBUFF_FILTER_NO_COOLDOWN
				and generaloptions.cooldown ~= DEBUFF_FILTER_COOLDOWN_ONLY)) then
			-- update button time every second or every tenth of a second
			--local timeRemaining = bt.expiretime - GetTime()
			--[[if timeRemaining < (DEBUFF_FILTER_TIMELEFT_TENTHSSEC + time_frame.intervalFractionSecond) then
				time_frame.timeInterval = time_frame.intervalFractionSecond
				time_frame.timeIntervalNotInTenths = false
			else
				time_frame.timeInterval = 1
				-- boolean value can be checked quickly instead of checking if the time interval is below 1 second
				time_frame.timeIntervalNotInTenths = true
			end]]
			
			DebuffFilter_SetTime(time_frame, bt.expiretime - GetTime());

			-- synchronize time so that its not displayed inbetween seconds
			--time_frame.update = time_frame.timeInterval - mod(timeRemaining, time_frame.timeInterval)
			--DebuffFilter_ScheduleUpdate(time_frame.updater, mod(timeRemaining, time_frame.timeInterval))

			if not time_frame.DBFisTimeVisible then
				--button:SetScript("OnUpdate",DebuffFilter_Button_OnUpdate);
				time_frame:Show()
				time_frame.DBFisTimeVisible = true
			end
		else
			hideTime(time_frame)
		end	
	else
		cooldown:Hide()
		hideTime(time_frame)

		-- dunno what this is for, maybe addon OmniCC?
		--if (button.timer) then
		--	button.timer:Hide();
		--end
	end


	button:Show();
end

--[[local FinishButtonTime = function(self)
	local time_frame = self:GetParent()
	if time_frame.DBFisTimeVisible then
		local remainTime = time_frame.expiretime-GetTime()

		if time_frame.timeIntervalNotInTenths and remainTime < (DEBUFF_FILTER_TIMELEFT_TENTHSSEC + time_frame.intervalFractionSecond) then 
			time_frame.timeInterval = time_frame.intervalFractionSecond
			time_frame.timeIntervalNotInTenths = false
		end
		DebuffFilter_SetTime(time_frame, remainTime);
		local nextUpdate = mod(remainTime, time_frame.timeInterval)
		if nextUpdate < 0.01 then nextUpdate = time_frame.timeInterval end
		if remainTime > 0 then DebuffFilter_ScheduleUpdate(self, nextUpdate) end
	end	
end]]

local UpdateAllEnabledFrames = function(self)
	for k, v in pairs(DebuffFilter.enabledFramesList) do
		for a, b in ipairs({"Buffs","Debuffs"}) do
			if v[b] then
				DebuffFilter_Frame_Update(k,b)
			end
		end
	end
end

--[[function DebuffFilter_ScheduleUpdate(updater, nextUpdate)
	if updater:IsPlaying() then
		updater:Stop()
	end
	updater:GetAnimations():SetDuration(nextUpdate)
	updater:Play()
end]]

function DebuffFilter:Sort(arr, len, comp_func)
	if len > 1 then
		if len > 10 then
			local start = DebuffFilter:merge_sort(arr, 1, len, comp_func)
			
			arr[1] = start
			for a = 2, len do
				arr[a] = start.next
				start = start.next
			end
		else
			local d, key
			for c = 1, (len-1) do
				d = c
				key = arr[c+1]
				while d > 0 and not comp_func(arr[d], key) do
					arr[d+1] = arr[d]
					d = d - 1
				end
				arr[d+1] = key
			end
		end
	end	
end

function DebuffFilter:merge_sort(arr, a, b, comp_func)
	local ret_val
	if (b - a) > 8 then
		local mid = floor((b+a)/2)
		local left = DebuffFilter:merge_sort(arr, a, mid, comp_func)
		local right = DebuffFilter:merge_sort(arr, mid+1, b, comp_func)
		ret_val = DebuffFilter:merge(left, right, comp_func)
	else
		ret_val = arr[a]
		local key, iter
		arr[a].next = nil
		for c = a, (b-1) do
			key = arr[c+1]
			iter = key
			iter.next = ret_val
			if comp_func(iter.next, key) then
				iter = iter.next
				while iter.next and comp_func(iter.next, key) do
					iter = iter.next
				end
				key.next = iter.next
				iter.next = key
			else
				ret_val = key
			end	
		end
	end
	return ret_val
end

function DebuffFilter:merge(left, right, comp_func)
	local start, iter
	if comp_func(left, right) then
		start = left
		left = left.next
	else
		start = right
		right = right.next
	end
	iter = start
	while left and right do
		if comp_func(left, right) then
			iter.next = left
			left = left.next
		else
			iter.next = right
			right = right.next
		end
		iter = iter.next
	end
	if left then
		iter.next = left
	else
		iter.next = right
	end
	iter = start
	return start
end

function DebuffFilter_Frame_Update(framename, buffOrDebuff)
	--if not DebuffFilter.PlayerConfig.Buffs[framename][buffOrDebuff].enabled then return end

	local frameitem = DebuffFilter.Frames[framename][buffOrDebuff];
	local framelayout = DebuffFilter.PlayerConfig.Frames[framename][buffOrDebuff];
	local buttons = frameitem.buttons;
	local frametarget = DebuffFilter.PlayerConfig.Frames[framename].frametarget
	-- these two tables are only used in this function, i save memory by not recreating them
	local already_seen_debuffs = frameitem.already_seen_debuffs;
	-- used to store filtered buffs/debuffs, so that they can be sorted
	local buff_table_array = frameitem.buff_table_array

	local button;
	local name, texture, applications, duration, expiretime, caster
	local nametexture;
	local width = 0;

	local debuff = nil
	local buff_table = nil
	-- determines if a (de)buff is in a group that user wants shown
	local buffInGroup
	-- under 5 seconds, period between updates of button time (from 0.1 to 1 second)
	local intervalFractionSecond = DEBUFF_FILTER_TIMELEFT_NUMFRACSEC
		
	local casterGUID, BB, maskedBB

	local i, isdebuff
	-- total number of buffs or debuffs for the frame's target (before filter)
	local numbuffs = 0
	local sb
	-- add buffs or debuffs to current frame (combining buffs and debuffs together) if user wants
	DebuffFilter.de_buffsList[1] = buffOrDebuff
	DebuffFilter.de_buffsList[2] = nil
	if DebuffFilter.PlayerConfig.Buffs[framename][buffOrDebuff].combineBuffsDebuffsFrames then
		DebuffFilter.de_buffsList[1] = "Buffs"
		DebuffFilter.de_buffsList[2] = "Debuffs"
	end
	-- iterate through every debuff/buff player has, filter it, and store correct ones in temporary array for sorting
	for _, de_buffs in ipairs(DebuffFilter.de_buffsList) do
		
		i = 1
		isdebuff = de_buffs == "Debuffs"
		sb = DebuffFilter.PlayerConfig.Buffs[framename][de_buffs]
		 
		buffInGroup = sb.showAllBuffs
		
		if isdebuff then
			name, _, texture, applications, frameitem.Comparisons["debufftype"], duration, expiretime, caster, frameitem.Comparisons["isStealable"] = UnitDebuff(frametarget, i);
		else
			name, _, texture, applications, frameitem.Comparisons["debufftype"], duration, expiretime, caster, frameitem.Comparisons["isStealable"] = UnitBuff(frametarget, i);
		end
		while texture do
			frameitem.Comparisons["ismine"] = caster == "player";

			if not buffInGroup and sb.showAllNonRaiderBuffs and caster ~= nil then
				casterGUID = UnitGUID(caster);
				if casterGUID ~= nil then 
					BB = tonumber(casterGUID:sub(5,5), 16);
					maskedBB = BB % 8; -- x % 8 has the same effect as x & 0x7 on numbers <= 0xf
					-- mask with 0x7 to get: 0 for players, 1 for world objects, 3 for NPCs, 4 for permanent pets, 5 for vehicles.
					if maskedBB ~= 0 and maskedBB ~= 4 and maskedBB ~= 5 then
						buffInGroup = true
					end
				end	
			end	

			-- check if (de)buff is in a group of debuffs/buffs that user selected
			if not buffInGroup then
				for k, v in pairs(DebuffFilter.Frames[framename][de_buffs].ComparisonList) do
					if frameitem.Comparisons[v.name] == v.value then
						buffInGroup = true
						break
					end	
				end
			end
			
			-- if the (de)buff is in a group, but its duration is too long, dont show it, unless the (de)buff
			-- was typed manually into the list
			if buffInGroup and sb.buffDurationToggle then
				if duration > sb.buffDuration or duration == 0 then
					buffInGroup = false
				end
			end
			
			--buffInGroup = sb.showAllBuffs or (sb.showAllMyBuffs and ismine) or (sb.showAllNonRaiderBuffs and isnonraider) or
			--	(sb.showAllStealableBuffs and isStealable) or (sb.showAllMagicBuffs and (debufftype == "Magic")) or
			--	(sb.showAllPoisonBuffs and (debufftype == "Poison")) or
			--	(sb.showAllDiseaseBuffs and (debufftype == "Disease")) or (sb.showAllCurseBuffs and (debufftype == "Curse"))
			debuff = nil
			-- check if current debuff/buff is on list of debuffs/buffs to display
			if not buffInGroup then
				debuff = DebuffFilter.PlayerConfig.Buffs[framename][de_buffs].List[name];
				-- don't show (de)buff unless its expiration is less than a certain period of time
				if debuff and sb.buffExpiretimeToggle and (expiretime > (GetTime() + sb.buffExpiretime) or expiretime == 0) then
					debuff = nil
				end
				if debuff and debuff.intervalFractionSecond then
					intervalFractionSecond = debuff.intervalFractionSecond
				end	
			else
				if sb.buffExpiretimeToggle and (expiretime > (GetTime() + sb.buffExpiretime) or expiretime == 0) then
					buffInGroup = false
				else
					intervalFractionSecond = DebuffFilter.PlayerConfig.Frames[framename][de_buffs].intervalFractionSecond
				end	
			end

			-- if texture field is not nil, look for a match with the name of debuff/buff's texture
			-- problem with this, is that i have seen spells who have texture names that are much different than the spell name
			if (debuff and (not debuff.selfapplied or frameitem.Comparisons["ismine"]) and (not debuff.texture or string.match(texture, debuff.texture))) 
					or buffInGroup then

				nametexture = name .. texture;
				if (already_seen_debuffs[nametexture] and not (debuff and debuff.dontcombine)) then
					-- below just ensures that buff with longer time span is shown
					buff_table = buff_table_array[already_seen_debuffs[nametexture]]
					if expiretime > buff_table.expiretime then
						buff_table.expiretime = expiretime
						buff_table.index=i
						buff_table.applications = applications
						buff_table.duration = duration
						buff_table.num_combines = buff_table.num_combines + 1
						buff_table.expiretime2 = expiretime
					end
				else
					width = width + 1

					buff_table = buff_table_array[width]
					if buff_table == nil then
						buff_table_array[width] = {
							name=name, texture=texture, applications=applications, duration=duration, expiretime=expiretime,
							index=i, num_combines=1, expiretime2 = expiretime, isdebuff = isdebuff, debufftype = frameitem.Comparisons["debufftype"],
							intervalFractionSecond=intervalFractionSecond
						}
						buff_table = buff_table_array[width]
					else
						buff_table.name = name
						buff_table.texture = texture
						buff_table.applications=applications
						buff_table.duration=duration
						buff_table.expiretime=expiretime
						buff_table.index=i
						buff_table.num_combines=1
						buff_table.expiretime2 = expiretime
						buff_table.isdebuff = isdebuff
						buff_table.debufftype = frameitem.Comparisons["debufftype"]
						buff_table.intervalFractionSecond = intervalFractionSecond
					end
					-- used for sorting, to keep sorting function minimal, 72000 seconds = 20 hours
					if buff_table.expiretime2 == 0 then buff_table.expiretime2 = 72000 end
					already_seen_debuffs[nametexture] = width;
				end
			end	

			i = i + 1;
			buffInGroup = sb.showAllBuffs
			if isdebuff then
				name, _, texture, applications, frameitem.Comparisons["debufftype"], duration, expiretime, caster, frameitem.Comparisons["isStealable"] = UnitDebuff(frametarget, i);
			else
				name, _, texture, applications, frameitem.Comparisons["debufftype"], duration, expiretime, caster, frameitem.Comparisons["isStealable"] = UnitBuff(frametarget, i);
			end
		end
		numbuffs = numbuffs + i - 1

		-- better to reuse this table then use up memory by creating a new one
		for k in pairs(already_seen_debuffs) do
			already_seen_debuffs[k] = nil;
		end
	end	

	-- display number of debuffs/buffs raider has
	if (width == 0) then
		frameitem.frameCount:SetText("");
	else
		frameitem.frameCount:SetText(numbuffs);
	end

	local sort_method = DEBUFF_FILTER_NO_SORT
	if framelayout.sort ~= DEBUFF_FILTER_NO_SORT then
		sort_method = framelayout.sort
	else
		if DebuffFilter.PlayerConfig.General.sort ~= DEBUFF_FILTER_NO_SORT then
			sort_method = DebuffFilter.PlayerConfig.General.sort
		end
	end	
	if sort_method ~= DEBUFF_FILTER_NO_SORT then
		if sort_method == DEBUFF_FILTER_SORT_ALPHA_ATOZ then
			DebuffFilter:Sort(buff_table_array, width, function(a,b) return a.name < b.name end)
		elseif sort_method == DEBUFF_FILTER_SORT_ALPHA_ZTOA then
			DebuffFilter:Sort(buff_table_array, width, function(a,b) return a.name > b.name end)
		elseif sort_method == DEBUFF_FILTER_SORT_TIMEREM_LONGTOSHORT then
			DebuffFilter:Sort(buff_table_array, width, function(a,b) return a.expiretime2 > b.expiretime2 end)
		elseif sort_method == DEBUFF_FILTER_SORT_TIMEREM_SHORTTOLONG then
			DebuffFilter:Sort(buff_table_array, width, function(a,b) return a.expiretime2 < b.expiretime2 end)
		end
	end	
	
	local enableMouse = not(DebuffFilter.PlayerConfig.General.lock and not DebuffFilter.PlayerConfig.General.tooltips)
	local bt, buttonname

	-- display buffs/debuffs for frame
	for j = 1, width do
		bt = buff_table_array[j]
		
		button = buttons["button" .. j];
		if (not button) then
			buttonname = "DebuffFilterButton"..framename..buffOrDebuff..j
			button = CreateFrame("Button", buttonname, frameitem.frame, "DebuffFilter_BuffButtonTemplate");
			buttons["button" .. j] = button
			-- needed by gametooltip for debuff/buff description
			button.target = frametarget;

			if DebuffFilter.PlayerConfig.General.noBoxAroundTime then
				button.time_frame.DBFframetexture:SetTexture(0,0,0,0)
			else
				button.time_frame.DBFframetexture:SetTexture(0,0,0,1)
			end
			
			--updater
			--local updater = button.time_frame:CreateAnimationGroup()
			--updater:SetLooping('NONE')
			--updater:SetScript('OnFinished', FinishButtonTime)

			--local aaa = updater:CreateAnimation('Animation'); 
			--aaa:SetOrder(1)
			--button.time_frame.updater = updater

			--button.time_frame.update = 0
			button.time_frame.DBFisTimeVisible = false;
			-- background of button's time is not resized unless the string lengh of the time changes
			button.time_frame.DBFtimeLen = 0
			button:EnableMouse(enableMouse);
			-- dont want ButtonFacade to use Count field, since it places it at bottom center of button,
			if LBF then 
				frameitem.buttonfacadeGroup:AddButton(button,{Count=false})
			end

			-- set the position of the button, and its time
			DebuffFilter_SetButtonLayout(framelayout, frameitem.frame, buttons, j);
		end

		DebuffFilter_ShowButton(button, bt, framename, buffOrDebuff);
	end

	-- hide remaining buttons that were visible from before
	i = width + 1;
	button = buttons["button" .. i];
	while button do
		button:Hide();
		i = i + 1;
		button = buttons["button" .. i];
	end
end

--[[function DebuffFilter:UNIT_AURA(eventName, ...)
	local arg1 = select(1, ...)
	local argGuid = UnitGUID(arg1)
	for k, v in pairs(DebuffFilter.Frames) do
		if UnitGUID(DebuffFilter.PlayerConfig.Frames[k].frametarget) == argGuid then
			DebuffFilter_Frame_Update(k, "Buffs")
			DebuffFilter_Frame_Update(k, "Debuffs")
		end
	end
end

function DebuffFilter:PLAYER_TARGET_CHANGED(eventName, ...)
	local target
	for k, v in pairs(DebuffFilter.Frames) do
		target = DebuffFilter.PlayerConfig.Frames[k].frametarget
		if target:sub(1,6) == "target" then
			-- if the unit the frame is on is not "target", the GUID will need to be updated so
			-- DebuffFilter_Frame_Update is not called twice
			v.Buffs.frame.targetGUID = UnitGUID(target)
			DebuffFilter_Frame_Update(k, "Buffs")
			v.Debuffs.frame.targetGUID = UnitGUID(target)
			DebuffFilter_Frame_Update(k, "Debuffs")
		end
	end
end

function DebuffFilter:PLAYER_FOCUS_CHANGED(eventName, ...)
	local target
	for k, v in pairs(DebuffFilter.Frames) do
		target = DebuffFilter.PlayerConfig.Frames[k].frametarget
		if target:sub(1,5) == "focus" then
			-- if the unit the frame is on is not "focus", the GUID will need to be updated so
			-- DebuffFilter_Frame_Update is not called twice
			v.Buffs.frame.targetGUID = UnitGUID(target)
			DebuffFilter_Frame_Update(k, "Buffs")
			v.Debuffs.frame.targetGUID = UnitGUID(target)
			DebuffFilter_Frame_Update(k, "Debuffs")
		end
	end
end]]

-- in combat, so show frames
function DebuffFilter:PLAYER_REGEN_DISABLED(eventName, ...)
	DebuffFilterFrame:Show();
end

-- not in combat, so hide frames
function DebuffFilter:PLAYER_REGEN_ENABLED(eventName, ...)
	DebuffFilterFrame:Hide();
end

-- update time on debuff/buff every second or tenth of a second; while loop is used to make sure, 
-- that if the players framerate is very low, it'll still do the updates as many times as needed
--[[function DebuffFilter_Button_OnUpdate(self, elapsed)

	self.update = self.update + elapsed;
	if (self.update >= self.timeInterval) then
		local timeRemaining = self.expiretime-GetTime()
		if timeRemaining > 0 then
			--self.update = self.update - self.timeInterval;
			if self.timeIntervalNotInTenths and timeRemaining < (DEBUFF_FILTER_TIMELEFT_TENTHSSEC + self.intervalFractionSecond) then 
				self.timeInterval = self.intervalFractionSecond
				self.timeIntervalNotInTenths = false
			end
			DebuffFilter_SetTime(self, timeRemaining);
			self.update = self.timeInterval - mod(timeRemaining, self.timeInterval)
		else
			self:GetParent():Hide()
		end
	end
end]]

-- update frame if the unit the frame is on has changed
--[[function DebuffFilter_Frame_OnUpdate(self, elapsed)

	self.DebuffFilterFrameUpdate = self.DebuffFilterFrameUpdate + elapsed;
	while (self.DebuffFilterFrameUpdate >= 0.1) do
		if self.targetGUID ~= UnitGUID(DebuffFilter.PlayerConfig.Frames[self.framename].frametarget) then
			self.targetGUID = UnitGUID(DebuffFilter.PlayerConfig.Frames[self.framename].frametarget)
			DebuffFilter_Frame_Update(self.framename,self.buffOrDebuff)
		end

		self.DebuffFilterFrameUpdate = self.DebuffFilterFrameUpdate - 0.1;
	end
end]]

-- update frame if the unit the frame is on has changed
--[[function DebuffFilter_Frame_Update_Every_FracSecond(self, elapsed)

	self.DebuffFilterFrameUpdateEveryFracSecond = self.DebuffFilterFrameUpdateEveryFracSecond + elapsed;
	if (self.DebuffFilterFrameUpdateEveryFracSecond >= 0.025) then
		DebuffFilter_Frame_Update(self.framename,self.buffOrDebuff)

		self.DebuffFilterFrameUpdateEveryFracSecond = 0.025 - mod(self.DebuffFilterFrameUpdateEveryFracSecond, 0.025);
	end
end]]


function DebuffFilter_SetTime(time_frame, time)
	local timeToShow

	if (time < DEBUFF_FILTER_TIMELEFT_TENTHSSEC) and time_frame.showFractionSecond then
		--local fracSec = (math.floor(time * 10 + 0.5)) / 10 - time_frame.intervalFractionSecond
		if time < 0 then
			timeToShow = "0.0"
		else
			-- truncate the hundredths of seconds and smaller, and make result a multiple of intervalFractionSecond
			local fracSec = time - mod(time, time_frame.intervalFractionSecond)
			timeToShow = tostring(fracSec)
			if timeToShow:len() == 1 then timeToShow = timeToShow .. ".0" end
		end	
	else
		local min, sec;

		-- avg time to nearest second or tenth of a second, and then subtract 1 or 0.1
		-- for example, with 28.1, we use 27, because the tenths of a second are decreasing until we hit 27.0,
		-- and there is a period of 1 second before the time is updated again
		--time = math.floor(time + 0.5) - 1;
		time = math.floor(time);
		if time < 0 then
			timeToShow = "0"
		else	
			if ( time >= 60 ) then
				min = math.floor(time/60);
				sec = time - min*60;
			else
				sec = time;
				min = 0;
			end
			timeToShow = tostring(sec)
			if min > 0 then
				if sec <= 9 then
					timeToShow = "0" .. timeToShow
				end
				timeToShow = min .. ":" .. timeToShow
			end
		end	
	end	

	local textcolor
	if (10 >= time) then
		textcolor = 0.27
	else
		textcolor = 0.82
	end
	if time_frame.DBFtextcolor ~= textcolor then
		time_frame.DBFtextcolor = textcolor
		time_frame.time:SetTextColor(1, textcolor, 0);
	end	

	if time_frame.DBFtimeToShow ~= timeToShow then
		time_frame.DBFtimeToShow = timeToShow
		time_frame.time:SetText(timeToShow)
	end	
	local timeLen = timeToShow:len()
	if time_frame.DBFtimeLen ~= timeLen then
		time_frame.DBFtimeLen = timeLen
		time_frame:SetWidth(time_frame.time:GetStringWidth()+4)
	end
end

-- reposition or redraw frame's debuffs/buffs after options have been changed
function DebuffFilter_UpdateLayout(framename,buffOrDebuff)
	local button;
	local frameitem = DebuffFilter.Frames[framename][buffOrDebuff];
	local layout = DebuffFilter.PlayerConfig.Frames[framename][buffOrDebuff];

	local i = 1;
	button = frameitem.buttons["button"..i];
	while button do
		button:ClearAllPoints();
		DebuffFilter_SetButtonLayout(layout, frameitem.frame, frameitem.buttons, i);

		i = i + 1;
		button = frameitem.buttons["button"..i];
	end
	
	DebuffFilter_SetCountOrientation(layout, frameitem);
end

-- set the location of a debuff/buff and the location of its time
function DebuffFilter_SetButtonLayout(layout, frame, buttons, index)
	local point, relpoint, x, y;
	local grow = layout.grow;
	local per_row = layout.per_row;
	local offset = 14;
	local button = buttons["button" .. index]

	point, relpoint = DebuffFilter.Orientation[grow].point, DebuffFilter.Orientation[grow].relpoint;
	x, y = DebuffFilter.Orientation[grow].x, DebuffFilter.Orientation[grow].y;

	-- place time to the left or right of debuff/buff if debuffs/buffs are arranged vertically
	if (per_row == 1 or layout.cooldown == DEBUFF_FILTER_COOLDOWN_ONLY
			or (layout.cooldown == DEBUFF_FILTER_NO_COOLDOWN and 
			DebuffFilter.PlayerConfig.General.cooldown == DEBUFF_FILTER_COOLDOWN_ONLY)) then
		offset = 4;
		DebuffFilter_SetTimeOrientation(layout.time_lr, button);
	else
		DebuffFilter_SetTimeOrientation(layout.time_tb, button);
	end

	if (index > 1) then
		-- start a new row if the current one has enough debuffs/buffs
		if (mod(index, per_row) == 1 or per_row == 1) then
			if (layout.grow == "rightdown" or layout.grow == "leftdown") then
				button:SetPoint("TOP", buttons["button" .. (index-per_row)], "BOTTOM", 0, -offset);
			else
				button:SetPoint("BOTTOM", buttons["button" .. (index-per_row)], "TOP", 0, offset);
			end
		else
			DebuffFilter_SetTimeOrientation(layout.time_tb, button);
			button:SetPoint(point, buttons["button" .. (index-1)], relpoint, x, y)
		end
	else
		button:SetPoint(point, frame, point, 0, 0);
	end
end

function DebuffFilter_SetTimeOrientation(orientation, button)
	local point, relpoint, x, y;

	point, relpoint = DebuffFilter.Orientation[orientation].point, DebuffFilter.Orientation[orientation].relpoint;
	x, y = DebuffFilter.Orientation[orientation].x, DebuffFilter.Orientation[orientation].y;

	-- set the position of the time, somewhere around the debuff/buff
	if (button.time_frame) then
		button.time_frame:ClearAllPoints();
		button.time_frame:SetPoint(point, button, relpoint, x, y);
	-- set the positions of the times for the debuffs/buffs of a certain frame
	else
		local time, buttonitem

		local i = 1;
		buttonitem = button["button"..i]
		while buttonitem do
			time = buttonitem.time_frame
			time:ClearAllPoints();
			time:SetPoint(point, buttonitem, relpoint, x, y);
			
			i = i + 1;
			buttonitem = button["button"..i]
		end
	end
end

-- set the location for the number of debuffs/buffs for the frame
function DebuffFilter_SetCountOrientation(layout, frameitem)
	local grow = layout.grow;
	local per_row = layout.per_row;
	local frame = frameitem.frame;

	local count = frameitem.frameCount;
	count:ClearAllPoints();

	if (per_row > 1) then
		if (grow == "rightdown" or grow == "rightup") then
			count:SetPoint("RIGHT", frame, "LEFT", 0, 0);
		else
			count:SetPoint("LEFT", frame, "RIGHT", 0, 0);
		end
	else
		if (grow == "rightdown" or grow == "leftdown") then
			count:SetPoint("BOTTOM", frame, "TOP", 0, 8);
		else
			count:SetPoint("TOP", frame, "BOTTOM", 0, -8);
		end
	end
end

-- lock the frames, so they cannot be moved, or unlock them
-- also allows mouse events to pass through debuffs/buffs if locked
function DebuffFilter_LockFrames(lock)
	local button, i, frameitem;

	for k, v in pairs(DebuffFilter.Frames) do
		for a, b in ipairs({"Buffs","Debuffs"}) do
			frameitem = DebuffFilter.Frames[k][b]
			frameitem.frame:EnableMouse(not lock);

			i = 1;
			button = frameitem.buttons["button" .. i];
			while button do
				button:EnableMouse(not lock);

				i = i + 1;
				button = frameitem.buttons["button" .. i];
			end
		end	
	end
end

-- code was moved from OnInitialize because otherwise, the saved positions of the frames could not be loaded.
-- These saved positions are only needed when user upgrades from old version of Debuff Filter.
-- These frames are from the xml file, and the saved positions are from the game -- the game saves the positions
-- of every xml frame when user logs out.
function DebuffFilter:VARIABLES_LOADED(eventName, ...)
	
	if DebuffFilter.portingXMLframes then
		local xmlFrameName, xmlframe, frameLayout
		for a, b in pairs({target="",focus="F",player="P"}) do
			for c, buff in ipairs({"Buffs","Debuffs"}) do
				xmlFrameName = "DebuffFilter_" .. b .. buff:sub(1,-2) .. "Frame"
				xmlframe = _G[xmlFrameName]
				frameLayout = DebuffFilter.PlayerConfig.Frames[a][buff]
				if xmlframe then 
					frameLayout.anchor = {"TOPLEFT", "UIParent", "BOTTOMLEFT", xmlframe:GetLeft(), xmlframe:GetTop()}
					DebuffFilter.Frames[a][buff].frame:ClearAllPoints();
					DebuffFilter.Frames[a][buff].frame:SetPoint(unpack(frameLayout.anchor));
				end	
			end
		end
	end	
end

function DebuffFilter:OnSkin(skin, gloss, backdrop, group, button, colors)
	if group then
		DebuffFilter.PlayerConfig.ButtonFacade[group] = {skin, gloss, backdrop, colors}
	end
end
	
-- ButtonFacade won't show entries for each frame unless you create its groups in OnEnable
function DebuffFilter:OnEnable()
	local groupID, frameitem, bfOptions
	for k, v in pairs(DebuffFilter.PlayerConfig.Frames) do
		for a, b in ipairs({"Buffs","Debuffs"}) do
			if LBF then
				groupID = k.." "..b.." Frame"
				frameitem = DebuffFilter.Frames[k][b]
				frameitem.buttonfacadeGroup = LBF:Group("DebuffFilter", groupID)
				bfOptions = DebuffFilter.PlayerConfig.ButtonFacade[groupID]
				if bfOptions then
					frameitem.buttonfacadeGroup:Skin(unpack(bfOptions))
				end
			end	

			-- display the frame and its debuffs/buffs
			--DebuffFilter_Frame_Update(k, b)
		end
	end

	--DebuffFilter:RegisterEvent("PLAYER_TARGET_CHANGED")
	--DebuffFilter:RegisterEvent("PLAYER_FOCUS_CHANGED")
	--DebuffFilter:RegisterEvent("UNIT_AURA")
end

function DebuffFilter:OnInitialize()
  	-- Code that you want to run when the addon is first loaded goes here.
	DebuffFilter:DebugTrace ("OnInitialize", true)
	-- create /dfilter command that enables changing options
	DebuffFilter:RegisterChatCommand("dfilter", "SlashProcessorFunc")
	-- options tables used to create options dialog
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("DebuffFilter", options)
    self.optionsFrame = {}
    -- options tables are added to the Blizzard Interface Options panel
    -- only thing I add is function to open standalone dialog, because the Buffs panel is only ok in standalone dialog
    self.optionsFrame.BlizOptions = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DebuffFilter", "DebuffFilter", nil, "BlizOptions")
    --[[self.optionsFrame.General = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DebuffFilter", "DebuffFilter", nil, "General")
    self.optionsFrame.Frames = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DebuffFilter", "Frame Layout","DebuffFilter", "Frames")
    self.optionsFrame.Buffs = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DebuffFilter", "Frame Buffs","DebuffFilter", "Buffs")]]
    -- configuration settings read from savedvariables folder
    self.db = LibStub("AceDB-3.0"):New("DebuffFilterDB", DebuffFilter:InitDefaults(), "Default")
	
	-- configuration for current character, as read from the savedvariables file
	DebuffFilter.PlayerConfig = self.db.char;

	-- used in update frame function if we are combining buffs and debuffs
	DebuffFilter.de_buffsList = {}
	-- iterate over list and display each frame every 0.025 seconds
	DebuffFilter.enabledFramesList = {}

	-- don't need this, DebuffTypeColor comes from somewhere, but dunno where
	--DebuffFilter.DebuffTypeColor = { Curse = {r=1, g=0, b=1}, Disease = {r=.55, g=.15, b=0},
	--	Magic = {r=0, g=0, b=1}, Poison = {r=0, g=1, b=0}, none = {r=1.0, g=0.7, b=0},
	--}

	-- create default frames in configuration if they don't exist
	for a, b in ipairs({"focus","target","player","targettarget"}) do
		if DebuffFilter.PlayerConfig.Frames[b] == nil then
			DebuffFilter.PlayerConfig.Frames[b] = {}
		end	
	end
	
	-- stores whether to grab the game's saved positions, of the xml frames from the old version of Debuff Filter
	DebuffFilter.portingXMLframes = false
	-- port settings from old version of Debuff Filter
	if DebuffFilter_Config then
		local DebuffFilter_Player = (UnitName("player").." - "..GetRealmName());
		local stri, ksub
		local buffOrDebuff, framename
		if DebuffFilter_Config[DebuffFilter_Player] then
			DebuffFilter.portingXMLframes = true
			for k, v in pairs(DebuffFilter_Config[DebuffFilter_Player]) do
				-- dont want to repeat code that checks what the frame name is, so i use ksub
				ksub = k
				if k:sub(1,4) == "all_" then
					ksub = k:sub(5)
				end

				buffOrDebuff = "Buffs"
				stri = ksub:find("debuff")
				if stri == nil then stri = ksub:find("buff") else buffOrDebuff = "Debuffs" end
				if stri == nil then
					DebuffFilter.PlayerConfig.General[k] = v
				else
					if stri == 1 then
						framename = "target"
					elseif stri == 2 and ksub:sub(1,1) == "p" then
						framename = "player"
					elseif stri == 2 and ksub:sub(1,1) == "f" then
						framename = "focus"
					end
					if k:sub(1,4) == "all_" then
						DebuffFilter.PlayerConfig.Buffs[framename][buffOrDebuff].showAllBuffs = v
					else
						if k:find("_layout") then
							for a, b in pairs(v) do
								DebuffFilter.PlayerConfig.Frames[framename][buffOrDebuff][a] = b
							end	
						elseif k:find("_list") then
							for a, b in pairs(v) do
								DebuffFilter.PlayerConfig.Buffs[framename][buffOrDebuff].List[a] = b
							end
						elseif k:sub(-1) == "s" then
							DebuffFilter.PlayerConfig.Buffs[framename][buffOrDebuff].enabled = v
						end
					end
				end
			end
			DebuffFilter_Config[DebuffFilter_Player] = nil
		end	
	end
	if DebuffFilter.PlayerConfig.General.cooldowncount == true then
		DebuffFilter.PlayerConfig.General.cooldowncount = false
		DebuffFilter.PlayerConfig.General.cooldown = DEBUFF_FILTER_COOLDOWN_ONLY
	end
			
	if LBF then
		-- buttonfacade callback
		LBF:RegisterSkinCallback("DebuffFilter", self.OnSkin, self)
	end

	-- holds all widget objects, such as the frames and buttons
	DebuffFilter.Frames = {}

	DebuffFilterFrame:SetScale(DebuffFilter.PlayerConfig.General.scale)

	-- position where frames start to be placed 
	DebuffFilter.xvalues = { xvalue = 478, nextxvalue = 578};
	DebuffFilter.yvalue = 335;
	for k, v in pairs(DebuffFilter.PlayerConfig.Frames) do
		-- name of the frame used to be its target also, but addon was updated so frames could be duplicated
		if v.frametarget == nil then v.frametarget = k end
		-- don't allow default frames to be listed, don't want to allow user to delete them
		if k ~= "player" and k ~= "target" and k ~= "focus" then
			options.args.General.args.listNewFrames.args[k] = newFrameName
		end
		-- create entries for frames in dropdown menus
		options.args.Frames.args[k] = frameLayoutTabs
		options.args.Buffs.args[k] = frameBuffsTabs
		DebuffFilter.Frames[k] = {}
		for a, b in ipairs({"Buffs","Debuffs"}) do
			DebuffFilter:createFrame(k, b)
		end
		DebuffFilter:adjustAnchorPoints()
	end

	if (DebuffFilter.PlayerConfig.General.lock and not DebuffFilter.PlayerConfig.General.tooltips) then
		DebuffFilter_LockFrames(true);
	end

	-- Update frames every 0.025 seconds
	local updater = DebuffFilterFrame:CreateAnimationGroup()
	updater:SetLooping('REPEAT')
	local anim = updater:CreateAnimation('Animation'); 
	anim:SetOrder(1)
	anim:SetScript('OnFinished', UpdateAllEnabledFrames)
	anim:SetDuration(0.025)
	updater:Play()

	if (DebuffFilter.PlayerConfig.General.combat) then
		DebuffFilter:RegisterEvent("PLAYER_REGEN_DISABLED");
		DebuffFilter:RegisterEvent("PLAYER_REGEN_ENABLED");

		if (not UnitAffectingCombat("player")) then
			DebuffFilterFrame:Hide();
		end
	end

	DebuffFilter:RegisterEvent("VARIABLES_LOADED")

	DebuffFilter:DebugTrace ("OnInitialize")
end

function DebuffFilter:createFrame(framename, buff)
	local frames = DebuffFilter.PlayerConfig.Frames
	local frameitem, frameLayout
	
	DebuffFilter.Frames[framename][buff] = {}
	frameitem = DebuffFilter.Frames[framename][buff]

	-- stores already seen debuffs/buffs so that they can be combined and stacked
	frameitem.already_seen_debuffs = {}
	-- stores (de)buffs that are to be displayed, so that they can be sorted
	frameitem.buff_table_array = {}
	
	frameitem.buttons = {}
	
	-- store only the (de)buff groups that user selected in comparisonlist, so that other comparisons
	-- can be skipped when filtering
	frameitem.Comparisons = {}
	frameitem.ComparisonList = {}
	local fc = frameitem.ComparisonList
	local sb = DebuffFilter.PlayerConfig.Buffs[framename][buff]
	if sb.showAllMyBuffs then fc.showAllMyBuffs = {name="ismine",value=true} end
	if sb.showAllStealableBuffs then fc.showAllStealableBuffs = {name="isStealable",value=1} end
	if sb.showAllMagicBuffs then fc.showAllMagicBuffs = {name="debufftype",value="Magic"} end
	if sb.showAllPoisonBuffs then fc.showAllPoisonBuffs = {name="debufftype",value="Poison"} end
	if sb.showAllDiseaseBuffs then fc.showAllDiseaseBuffs = {name="debufftype",value="Disease"} end
	if sb.showAllCurseBuffs then fc.showAllCurseBuffs = {name="debufftype",value="Curse"} end
	
	frameitem.frame = CreateFrame("Frame","DebuffFilter"..framename..buff,nil,"DebuffFilter_FrameTemplate");
	-- needed so that when a button is clicked, I can grab the frame's name and buff, so I can get layout information
	frameitem.frame.DBFframename = framename
	frameitem.frame.DBFbuffOrDebuff = buff
	frameitem.frameBackdrop = CreateFrame("Button",nil,frameitem.frame,"DebuffFilter_BackdropTemplate");
	frameitem.frameBackdropTitle = frameitem.frameBackdrop:CreateFontString(nil,"ARTWORK","DebuffFilter_NameTemplate");
	frameitem.frameBackdropTitle:SetText(framename .. " " ..buff);

	frameitem.frameCount = _G["DebuffFilter"..framename..buff.."Count"];

	frameLayout = frames[framename][buff]
	
	frameitem.frame:SetScale(frameLayout.scale)
	if frameLayout.anchor == nil then
		frameLayout.anchor = {"TOPLEFT","UIParent","BOTTOMLEFT",DebuffFilter.xvalues.xvalue,DebuffFilter.yvalue}
		
		local parentMaxY = DebuffFilterFrame:GetParent():GetTop() / DebuffFilter.PlayerConfig.General.scale
		local parentMaxX = DebuffFilterFrame:GetParent():GetRight() / DebuffFilter.PlayerConfig.General.scale
		local frameWidth = frameitem.frame:GetWidth()
		if (DebuffFilter.xvalues.xvalue + frameWidth) > parentMaxX then
			frameLayout.anchor[4] = parentMaxX - frameWidth
		end
		if DebuffFilter.yvalue > parentMaxY then
			frameLayout.anchor[5] = parentMaxY
		end
	end	

	frameitem.frame:ClearAllPoints();
	frameitem.frame:SetPoint(unpack(frameLayout.anchor));
	DebuffFilter:swapXvalues()

	-- hide a frame that the player unchecked
	if (not DebuffFilter.PlayerConfig.Buffs[framename][buff].enabled) then
		frameitem.frame:Hide();
	else
		if not DebuffFilter.enabledFramesList[framename] then
			DebuffFilter.enabledFramesList[framename] = {}
		end
		DebuffFilter.enabledFramesList[framename][buff] = true
	end

	if (DebuffFilter.PlayerConfig.General.backdrop) then
		frameitem.frameBackdrop:Show()
	end

	-- number of debuffs/buffs for a frame, including those not shown
	if (DebuffFilter.PlayerConfig.General.count) then
		frameitem.frameCount:Show();
	end

	-- position frameCount
	DebuffFilter_SetCountOrientation(frameLayout, frameitem);
	
	--frameitem.frame.DebuffFilterFrameUpdateEveryFracSecond = 0
	--frameitem.frame:SetScript("OnUpdate",DebuffFilter_Frame_Update_Every_FracSecond)
	
	-- for targets other than these default ones, I cannot tell if the unit, that the frame is on, has changed.
	-- So I needed to create an OnUpdate script for these frames.
	--[[local target = frames[framename].frametarget
	if target ~= "player" and target ~= "target" and target ~= "focus" then
		frameitem.frame.DebuffFilterFrameUpdate = 0
		frameitem.frame.targetGUID = UnitGUID(target)
		frameitem.frame:SetScript("OnUpdate",DebuffFilter_Frame_OnUpdate)
	end]]
end

function DebuffFilter:SlashProcessorFunc(input)
	DebuffFilter:DebugTrace ("SlashProcessorFunc", true)
  	-- Process the slash command ('input' contains whatever follows the slash command)
	if not input or input:trim() == "" then
		-- open addon's options in the Blizzard Interface Options panel
		--InterfaceOptionsFrame_OpenToCategory(self.optionsFrame.General)
		DebuffFilter:OpenConfigDialog()
	elseif input == "help" then
		DEFAULT_CHAT_FRAME:AddMessage("Debuff Filter commands:");
		DEFAULT_CHAT_FRAME:AddMessage("/dfilter: display the configuration menu in Blizzard's Interface Options panel.");
		DEFAULT_CHAT_FRAME:AddMessage("/dfilter |cff00ccffdialog|r: display configuration menu in a standalone panel.");
		DEFAULT_CHAT_FRAME:AddMessage("To move the frames, shift+left click and drag a backdrop or a monitored debuff/buff.");
		DEFAULT_CHAT_FRAME:AddMessage("To change a time orientation, ctrl+right click.");
	end
	DebuffFilter:DebugTrace ("SlashProcessorFunc")
end

-- this is a standalone dialog, it can be resized and moved, so user can see his config changes
function DebuffFilter:OpenConfigDialog()
	InterfaceOptionsFrame:Hide()
	-- BlizOptions are only shown through Blizzard's Interface panel, it is used to open this standalone dialog
	options.args.BlizOptions.hidden = true
	DebuffFilter.Container = LibStub("AceGUI-3.0"):Create("Frame")
	-- gotta create a new container, and then release it always, otherwise a bug can occur where you
	-- close, but the execute buttons are still visible, and still work!  (like delete)
	DebuffFilter.Container:SetCallback("OnClose",function(widget) options.args.BlizOptions.hidden = false 
			LibStub("AceGUI-3.0"):Release(DebuffFilter.Container)
		end)
	LibStub("AceConfigDialog-3.0"):Open("DebuffFilter",DebuffFilter.Container)
end